#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

# twbot.rb: Twitter Bot Support Library
# Version 0.15
# (C)2008- H.Hiro(Maraigue) http://hhiro.net/about/
# 
# Twitterのbot、特に「定期的にデータを取得→定期的に取得したデータを投稿」
# するタイプのbotの作成を支援するライブラリです。
# 
# 複製・改変の条件は（新）BSDライセンスに合わせます。
# 末尾の著作権表示をご覧下さい。

require 'net/http'
require 'yaml'
require 'rexml/document'
require 'cgi'

$KCODE = 'u' if RUBY_VERSION < "1.9.0"

class TwBot
  # ========================================
  #   内部で利用する関数
  # ========================================
  
  # ログファイルに書き込む
  def output_log(outmsg)
    STDERR.puts outmsg
    @log.puts outmsg if @log
  end
  
  # 設定ファイルに書き込む
  def output_config(config_file)
    return unless @write_config
    
    # YAML.dumpでこける場合に備え、openの前にdumpしている
    buf = YAML.dump(@config)
    open(config_file, 'w'){ |file| file.print buf }
  end
  
  # XMLが、先頭に指定された名前の要素を持っているか判定
  # (例) TwBot.xml_begin_with("<foo><bar></bar></foo>", "foo") はtrue
  #      TwBot.xml_begin_with("<foo><bar></bar></foo>", "bar") はfalse
  # xmlstr にはXMLのテキストを指定
  # XMLが不正な場合はnilを返す
  def TwBot.xml_begin_with(xmlstr, name)
    begin
      REXML::Document.new(xmlstr).elements.to_a("/#{name}").size > 0
    rescue REXML::ParseException
      nil
    end
  end
  
  # 文字列がtrue, falseのいずれかなら、それを該当するデータに変換して返す。
  # そうでなければ例外。
  def TwBot.parse_boolean(str)
    case str
    when /\Atrue\z/i
      true
    when /\Afalse\z/i
      false
    else
      raise ArgumentError, "Value is neither of 'true' nor 'false'"
    end
  end
  
  # ========================================
  #   インスタンス関数
  # ========================================
  
  # 継承先にて、データを取得するためのコードを記述する。
  # 返り値は文字列の配列でなければならない。
  def load_data
    raise NotImplementedError, "Method 'load_data' must be implemented"
  end
  private :load_data
  
  # 発言の投稿を行う
  def post_data
    # 投稿する発言が残っていない場合
    if @config[@list].empty?
      @logmsg = "(No post remain)"
      return nil
    end
    
    # 投稿する
    post = @config[@list].first
    
    if post.kind_of?(Array)
      reply_status = "&in_reply_to_status_id=#{post[1]}"
      post = post[0]
    else
      reply_status = ""
    end
    
    if post.empty?
      @logmsg = "(Empty string specified. Posting canceled.)"
      @config[@list].shift
      return false
    end
    if post == @config["last_post/"]
      post += " ."
    end
    
    res = nil
    
    http = Net::HTTP.new('twitter.com')
    unless @testmode
      http.start do |w|
        req = Net::HTTP::Post.new('/statuses/update.xml')
        req.basic_auth(@config['login/'], @config['password/'])
        res = w.request(req, 'source=twbotrb&status='+CGI.escape(post)+reply_status)
      end
    else
      res = Struct.new(:body).new('<status></status>') # dummy
    end
    
    # 正常に投稿されたか確認する
    # （正常に投稿されたとき…先頭の要素として"<status>"が含まれている）
    unless TwBot.xml_begin_with(res.body, "status")
      raise RuntimeError, "Posting a tweet has failed - XML data is:\n#{res.body}"
    end
    xml = REXML::Document.new(res.body)
    
    # 結果とログを出力
    STDERR.puts "[Updated!#{@testmode ? '(test)' : ''}] #{post} #{reply_status}"
    @config[@list].shift
    @config["last_post/"] = post
    @logmsg = "(A tweet has posted)"
    true
  end
  private :post_data
  
  # コンストラクタ
  # mode:
  #   動作("load"あるいは"post")を指定します。
  # config_file:
  #   設定ファイルです。YAML形式で保存されます。
  #   以下のキーを含むハッシュでなければなりません。
  #   "login/":ログイン名
  #   "password/":パスワード
  # log_file:（省略可能）
  #   ログを出力するファイルです。
  #   指定しないかnilを指定すると、ログのファイルへの出力を行いません。
  # list_name:（省略可能）
  #   1つの設定ファイルには複数のリストを持たせることが出来ます。
  #   mode=="load"の場合、ここで指定されたリストに発言の内容が格納されます。
  #   mode=="post"の場合、ここで指定されたリストの発言のうち先頭にある発言が
  #   投稿されます。
  # keep_config:（省略可能）
  #   trueの場合、設定ファイルの更新を行いません。
  #   なおfalseが指定されても、処理に失敗した場合（発言取得および投稿において
  #   例外が発生した場合）は設定ファイルの更新は行いません。
  # testmode: （省略可能、mode == "post"のときのみ有効)
  #   trueの場合、実際の発言の投稿を行いません。
  def initialize(mode, config_file, log_file = nil, list_name = '',
                 keep_config = false, testmode = false)
    @list = "data/#{list_name}"
    
    @config = YAML.load_file(config_file)
    @config['login/'] ||= ''
    @config['password/'] ||= ''
    @config[@list] ||= []
    
    @log = log_file ? open(log_file, 'a') : nil
    
    @logmsg = '(No error)' # この値がログファイルに書き込まれる
    @write_config = !keep_config
    @testmode = testmode
    
    # modeに応じて処理
    outmsg = nil
    begin
      case mode
      when 'load'
        # load(発言内容の生成)
        @config[@list].concat(load_data())
        
        output_log("[#{Time.now}]<#{mode}> #{@logmsg}")
        output_config(config_file)
      when /\Apost(?:=(\d+)(?:,(\d+))?)?\z/
        # post(発言の投稿)
        # "post" or "post=AAA" or "post=AAA,BBB" の形式で指定
        # (AAA: 投稿回数 BBB: 投稿失敗時にリトライする回数)
        post_count = $1 || 1
        retry_count = $2 || 1
        post_count = post_count.to_i
        retry_count = retry_count.to_i
        
        if post_count <= 0
          raise ArgumentError, "Posting count must be positive (#{post_count} specified)"
        end
        if retry_count <= 0
          raise ArgumentError, "Retrying count must be positive (#{retry_count} specified)"
        end
        
        STDERR.puts "Post for #{post_count} times / Quit with #{retry_count} failure"
        
        # 投稿する
        post_count.times do
          config_orig = @config.dup
          post_result = false
          
          begin
            post_result = post_data()
          rescue Exception => e
            output_log("[#{Time.now}]<#{mode}> #{e}: #{e.backtrace.join(%'\n')}")
            @config = config_orig
            
            retry_count -= 1
            break if retry_count <= 0
            redo
          else
            output_log("[#{Time.now}]<#{mode}> #{@logmsg}")
            output_config(config_file)
          end
          
          break if post_result == nil # if no message to be posted remains
        end
      else
        # modeの指定が不正な場合
        raise ArgumentError, 'Not a valid mode'
      end
    rescue Exception => e
      output_log("[#{Time.now}]<#{mode}> #{e}: #{e.backtrace.join(%'\n')}")
    end
    @log.close if @log
  end
  
  # ========================================
  #   ユーティリティ関数
  # ========================================
  
  # 文字列からすべての"@USERNAME"を除去し、"@ USERNAME"にしたものを返す。
  # ブロックを渡した場合、ユーザ名(先頭の@除く)をブロックに与え、
  # その結果が真であった場合のみ"@USERNAME"を除去する。
  def TwBot.remove_reply(str)
    str.gsub(/(@|＠)([0-9A-Z_a-z]+)/) do |x|
      if block_given?
        yield($2) ? "#{$1} #{$2}" : x
      else
        "#{$1} #{$2}"
      end
    end
  end
  
  # ページに分けて結果を返すAPIのうち、返ってくるデータがユーザで
  # あるものを取得し、そのユーザ名一覧を返す。
  # 
  # APIによる取得が retry_count 回を超えたら強制終了する。
  # 返り値はハッシュで、
  # {:result => ["user1", "user2", ...],
  #  :error => [exception1, exception2, ...]}
  # という形をしている。
  # ただし、取得に失敗した場合は :result に対応する値は nil である。
  def TwBot.retrieve_paginated_user_list(path, login, pass, retry_count)
    user_list = []
    error_list = []
    cursor = -1
    
    while true
      begin
        STDERR.puts "Downloading #{path} (cursor ID: #{cursor})"
        
        # Download followers
        xml_source = nil
        Net::HTTP.start('twitter.com') do |http|
          req = Net::HTTP::Get.new(path)
          req.basic_auth login, pass
          res = http.request(req, "cursor=#{cursor}")
          res.value
          xml_source = res.body
        end
        
        # Parse XML
        xml = REXML::Document.new(xml_source)
        user_list.concat(
          xml.elements.to_a("/users_list/users/user/screen_name").map{
            |user_info| user_info.text } )
        
        cursor = xml.elements.to_a("/users_list/next_cursor").first.text
        break if cursor == "0"
      rescue Exception => e
        STDERR.puts e
        STDERR.puts e.backtrace
        
        error_list << e
        retry_count -= 1
        break if retry_count <= 0
      end
    end
    
    {:result => (retry_count <= 0 ? nil : user_list.uniq), :error => error_list}
  end
  
  # follower一覧を返す。
  # 詳細は get_paginated_user_list を参照。
  def TwBot.retrieve_followers(login, pass, retry_count = 3)
    TwBot.retrieve_paginated_user_list('/statuses/followers.xml', login, pass, retry_count)
  end
  
  # friend一覧を返す。
  # 詳細は get_paginated_user_list を参照。
  def TwBot.retrieve_friends(login, pass, retry_count = 3)
    TwBot.retrieve_paginated_user_list('/statuses/friends.xml', login, pass, retry_count)
  end
  
  # 上記2メソッドのラッパー。
  # load_dataメソッド内でこれらを呼び出すと、現在指定されている
  # ユーザ名・パスワードでfollowers/friendsを取得する。
  def get_followers
    TwBot.retrieve_followers(@config['login/'], @config['password/'])
  end
  
  def get_friends
    TwBot.retrieve_friends(@config['login/'], @config['password/'])
  end
  
  # ユーザを指定してフォローする。
  # 成功したらtrue, 失敗したら例外
  def TwBot.make_following(login, pass, user)
    res = nil
    Net::HTTP.start('twitter.com') do |http|
      req = Net::HTTP::Post.new("/friendships/create/#{user}.xml")
      req.basic_auth(login, pass)
      res = http.request(req)
      res.value
    end
    
    (TwBot.xml_begin_with(res.body, "user") ||
     raise(RuntimeError, "Invalid XML for following status: #{res.body}"))
  end
  
  # ユーザを指定してフォローを外す。
  # 成功したらtrue, 失敗したら例外
  def TwBot.make_unfollowing(login, pass, user)
    res = nil
    Net::HTTP.start('twitter.com') do |http|
      req = Net::HTTP::Post.new("/friendships/destroy/#{user}.xml")
      req.basic_auth(login, pass)
      res = http.request(req)
      res.value
    end
    
    (TwBot.xml_begin_with(res.body, "user") ||
     raise(RuntimeError, "Invalid XML for following status: #{res.body}"))
  end
  
  # login のユーザが target のユーザをフォローしているか/されているか調べる。
  # 成功した場合、{:following => true, :followed => false} のようなハッシュを返す。
  # （:following - login のユーザが target のユーザをフォローしている）
  # （:followed  - login のユーザが target のユーザにフォローされている）
  # 失敗したら例外
  def TwBot.check_following(login, pass, target)
    res = nil
    xml_source = nil
    
    # Download
    Net::HTTP.start('twitter.com') do |http|
      req = Net::HTTP::Get.new("/friendships/show.xml")
      req.basic_auth(login, pass)
      res = http.request(req, "source_screen_name=#{login}&target_screen_name=#{target}")
      res.value
      xml_source = res.body
    end
    
    # parse the XML
    xml = REXML::Document.new(xml_source)
    {:following =>
       TwBot.parse_boolean(
         xml.elements.to_a("/relationship/source/following").first.text),
     :followed =>
       TwBot.parse_boolean(
         xml.elements.to_a("/relationship/source/followed_by").first.text)}
  end
  
  # 上記3メソッドのラッパー。
  # load_dataメソッド内でこれらを呼び出すと、現在指定されている
  # ユーザ名・パスワードでfollowers/friendsを取得し、あるいは
  # フォローしているかをチェックする。
  def follow(user)
    TwBot.make_following(@config['login/'], @config['password/'], user)
  end
  
  def unfollow(user)
    TwBot.make_unfollowing(@config['login/'], @config['password/'], user)
  end

  def following_status(target)
    TwBot.check_following(@config['login/'], @config['password/'], target)
  end
end

# Copyright (c) 2008- H.Hiro(Maraigue)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
# Neither the name of the Maraigue nor the names of its contributors
# may be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
